(*  Title:      ISM/AmbISM.thy
    Id:         $Id: AmbISM.thy,v 1.5 2003/05/28 11:40:06 dvo Exp $
    Author:     David von Oheimb
    License:    (C) 2002 Siemens AG; any non-commercial use is granted

*)

header {* Ambient ISMs *}

theory AmbISM
imports ISM
begin

record ('i, 'a) astate = 
  parent :: "'a ~=> 'a"  --{* ambient forest, intended to be acyclic (but this is not necessary) *}
  home   :: "'i ~=> 'a"  --{* assignment of ISMs to their home ambients *}
      
datatype ('i, 'a) acmd  --{* also known as capabilities *}

= Assign 'i 'a --{* assign an AmbISM to the current ambient *}
| In  'a  --{* move into an ambient   *}
| Out 'a  --{* move out of an ambient *}
| Del 'a  --{* delete (open) an ambient *}
| Ins 'a "'a set" "'i set"  --{* insert an ambient, moving specified subambients and AmbISMs into it *}

consts
  astate_trans :: "('i, ('i,'a) acmd, ('i,'a) astate) g_trans"
(*<*)
syntax (xsymbols)
"@astate_trans" :: "('i, 'a) astate \<Rightarrow> 'i \<Rightarrow> ('i,'a) acmd \<Rightarrow> 
                    ('i, 'a) astate \<Rightarrow> bool" ("_ \<midarrow>_\<Colon>_\<rightarrow> _" [61,61,61,61] 60)
"@astate_trans*":: "('i, 'a) astate \<Rightarrow> 'i \<Rightarrow> ('i,'a) acmd list \<Rightarrow> 
                    ('i, 'a) astate \<Rightarrow> bool" ("_ \<midarrow>_\<Colon>_\<rightarrow>* _" [61,61,61,61] 60)
translations
  "as \<midarrow>i\<Colon>c \<rightarrow>  as' " \<rightleftharpoons> "(as, c , as') \<in>           astate_trans i"
  "as \<midarrow>i\<Colon>cs\<rightarrow>* as' " \<rightleftharpoons> "(as, cs, as') \<in> fold_rel (astate_trans i)"
(*>*)
inductive "astate_trans i" intros
Assign: "home ai i = Some m \<Longrightarrow> 
      n = m \<or> parent ai n = Some m \<Longrightarrow> home ai j = None \<Longrightarrow> 
                         ai \<midarrow>i\<Colon>Assign j n\<rightarrow> ai\<lparr>home:=(home ai)(j\<mapsto>n)\<rparr>"
In:  "home ai i = Some m \<Longrightarrow> n \<noteq> m \<Longrightarrow> parent ai m = parent ai n \<Longrightarrow> 
                         ai \<midarrow>i\<Colon>In  n\<rightarrow> ai\<lparr>parent:=(parent ai)(m\<mapsto>n)\<rparr>"
Out: "home ai i = Some m \<Longrightarrow> n \<noteq> m \<Longrightarrow> parent ai m = Some n \<Longrightarrow> 
                         ai \<midarrow>i\<Colon>Out n\<rightarrow> ai\<lparr>parent:=(parent ai)(m:=parent ai n)\<rparr>"
Del: "home ai i \<in> {Some n, Some m} \<Longrightarrow> n \<noteq> m \<Longrightarrow> parent ai n = Some m \<Longrightarrow> 
                         ai \<midarrow>i\<Colon>Del n\<rightarrow> ai\<lparr>parent:=((parent ai)(n\<leadsto>m))(n:=None),
                                          \<spacespace>home  :=(home   ai)(n\<leadsto>m)\<rparr>"
Ins: "home ai i = Some m \<Longrightarrow> n \<notin> dom (parent ai) \<union> ran (parent ai) \<union> ran (home ai) \<Longrightarrow> 
                       parent ai`ns \<subseteq> {Some m} \<Longrightarrow> home ai`is \<subseteq> {Some m} \<Longrightarrow> 
                      ai \<midarrow>i\<Colon>Ins n ns is\<rightarrow> ai\<lparr>parent:=(parent ai)(ns{\<mapsto>}n)(n\<mapsto>m),
                                            \<spacespace>home  :=(home   ai)(is{\<mapsto>}n)\<rparr>"
(*<*)
declare astate_trans.intros [intro!]
lemma astate_trans_Ins:  --{*implicit from premises: @{term "n \<noteq> m \<and> n \<notin> ns"}*}
  "n \<notin> dom (parent ai) \<Longrightarrow> n \<notin> ran (parent ai) \<Longrightarrow> n \<notin> ran (home ai) \<Longrightarrow> 
   home ai i = Some m \<Longrightarrow> parent ai`ns \<subseteq> {Some m} \<Longrightarrow> home ai`is \<subseteq> {Some m} \<Longrightarrow>
    ai \<midarrow>i\<Colon>Ins n ns is\<rightarrow> 
    ai\<lparr>parent:=(parent ai)(ns{\<mapsto>}n)(n\<mapsto>m), home:=(home ai)(is{\<mapsto>}n)\<rparr>"
by (fast)
declare astate_trans.Ins[rule del] astate_trans_Ins[intro!]
inductive_cases astate_trans_elim_cases [elim!]: 
   "ai \<midarrow>i\<Colon>Assign j n \<rightarrow> ai'"
   "ai \<midarrow>i\<Colon>In  n      \<rightarrow> ai'"
   "ai \<midarrow>i\<Colon>Out n      \<rightarrow> ai'"
   "ai \<midarrow>i\<Colon>Del n      \<rightarrow> ai'"
   "ai \<midarrow>i\<Colon>Ins n ns is\<rightarrow> ai'"

lemma In_Out: "home ai i = Some m \<Longrightarrow> parent ai m = parent ai n \<Longrightarrow> n \<noteq> m \<Longrightarrow> 
 (ai, [In n, Out n], ai') \<in> fold_rel (astate_trans i) = (ai = ai')"
apply (auto)
apply (erule subst)
apply (simp)
done
lemma Out_In: "home ai i = Some m \<Longrightarrow> parent ai m = Some n \<Longrightarrow> n \<noteq> m \<Longrightarrow> 
 (ai, [Out n, In n], ai') \<in> fold_rel (astate_trans i) = (ai = ai')"
apply (auto)
apply (erule subst)
apply (simp)
done

lemma astate_eqI [rule_format]: 
 "parent ai = parent ai' \<longrightarrow> home ai = home ai'\<longrightarrow> ai = (ai'::('i, 'a) astate)"
by (induct "ai", simp)

lemma Ins_Del: 
 "n \<notin> dom (parent ai) \<union> ran (parent ai) \<union> ran (home ai) \<Longrightarrow> 
  home ai i = Some m \<Longrightarrow> parent ai `ns \<subseteq> {Some m} \<Longrightarrow> 
  home ai `is \<subseteq> {Some m} \<Longrightarrow> 
  (ai, [Ins n ns is, Del n], ai') \<in> fold_rel (astate_trans i) = (ai = ai')"
apply (safe del: disjE, erule swap, rule astate_eqI, auto)
apply     (rule ext, simp split add: split_if_asm, fast intro: domI ranI)+
apply (rule_tac t = "ai'" in subst, rule_tac [2] astate_trans.Del,
       auto intro: domI ranI)
apply (rule astate_eqI, auto, rule_tac [!] ext, 
       auto intro: ranI split add: split_if_asm simp add: dom_def)
done

lemma Del_Ins: "home ai i \<in> {Some n, Some m} \<Longrightarrow> n \<noteq> m \<Longrightarrow> parent ai n = Some m \<Longrightarrow> 
 (ai, [Del n, Ins n {n'. parent ai n' = Some n} {i. home ai i = Some n}], ai') 
  \<in> fold_rel (astate_trans i) = (ai = ai')"
apply (safe del: disjE, erule swap, rule astate_eqI, auto del: disjE)
apply     (rule ext, clarsimp)+
apply   (rule_tac t = "ai'" in subst, rule_tac [2] astate_trans_Ins, 
          auto simp: ran_def)
prefer 3
apply   (rule_tac t = "ai'" in subst, rule_tac [2] astate_trans_Ins, 
          auto simp add: ran_def)
apply   (erule_tac [3] swap, rule_tac [!] astate_eqI)
apply      (auto, rule_tac [!] ext, auto split add: split_if_asm)
done
(*>*)

constdefs
  Amb_trans :: "('i, ('i, 'a) acmd list, ('i, 'a) astate) g_trans"
 "Amb_trans \<equiv> \<lambda>i. {(ai,acmds,ai'). i \<in> dom (home ai) \<and> ai \<midarrow>i\<Colon>acmds\<rightarrow>* ai'}"

  vicinity :: "('i, 'a) astate \<Rightarrow> 'i \<Rightarrow> 'a set" --{* home ambient plus its parent *}
 "vicinity ai i \<equiv> case home ai i of None \<Rightarrow> {} | Some n \<Rightarrow> {n} \<union> o2s (parent ai n)"

 "local" :: "('i, 'a) astate \<Rightarrow> 'i \<Rightarrow> 'i set"
 "local ai i \<equiv> {j. vicinity ai i \<inter> vicinity ai j \<noteq> {}}"

  local_output :: "('i, ('c, 'p, 'm, 's) ism) family \<Rightarrow> ('i, 'a) astate \<Rightarrow>
                   ('i, ('c, 'p, 'm, 's) ism) family"
 "local_output \<equiv> \<lambda>(A,I) ai. 
      (\<lambda>i. (A i)\<lparr>outputs := outputs (A i) \<inter> (\<Union>j\<in>local ai i. inputs (A i))\<rparr>,I)"

  Amb_comp_runs :: "('i, (('i, 'a) acmd list, 'p, 'm, 's) ism) family \<Rightarrow> 
    ('i, 'a) astate \<Rightarrow> ((('p, 'm, ('i, 'a) astate \<times> ('i \<Rightarrow> 's)) conf) list) set"
 "Amb_comp_runs AI ai0 \<equiv> comp_runs (local_output AI) ai0 Amb_trans"
--{* only local output is allowed *}
--{* all AmbISMs are running; port ownership is statically fixed; ports are constantly enabled *}

(*<*)
lemma wf_comp_local_output_eq [simp]:
  "\<And>AI. wf_comp (local_output AI ai0) = wf_comp AI"
by (force simp add: local_output_def wf_comp_def)
(*>*)
lemma Amb_comp_runs_preserves_wf_comp: 
 "wf_comp AI \<Longrightarrow> cs \<in> Amb_comp_runs AI ai0 \<Longrightarrow> 
  \<forall>(b,ai,s) \<in> set cs. wf_comp (local_output AI ai)"
(*<*)
by (unfold Amb_comp_runs_def, erule comp_runs_global_induct, simp_all)
(*>*)

end
